

import matplotlib as mpl
import matplotlib.pyplot as plt

from typing import List
from pathlib import Path

from .appearance import *

from .classify import *

from .plotter import *

from .facets import *

from .chart import *

from .precast import *

# TODO: some alpha fails for vspan when using daspi and ggplot2 styles
style.use('daspi')

